{*******************************************************************************
 Project       : "Space Shooter ..." ?
 Date          : 2007-07-11
 Version       : 1.3
 Compiler Ver. : Delphi 2007
 Author        : Denny Koberling
 E-Mail        : JasonVoorhees2k@arcor.de
 Copyright     : Copyright (c) 2007 Denny Koberling
*******************************************************************************}

unit USprite;

interface

uses Windows, Graphics, SysUtils, Math, UAnimBmp, UTimer, UType;

type
  TSprite = class(TObject)
  private
    fLayer        : TLayer;
    fParent       : TSprite;
    fAnimBmp      : TAnimBmp;
    fX, fY        : Single;
    fSpeed        : Single;
    fDead         : Boolean;
    fFireTimer    : TTimer;
    fAnimTimer    : TTimer;
    fAnimLoop     : Boolean;
    fAnimNr       : Byte;
    fAnimDefaultNr: Byte;
    fFrameNr      : Byte;
    procedure SetAnimNr(aAnimNr: Byte);
    function GetCenterX: Single;
    function GetCenterY: Single;
    function GetPixel(aX, aY: Integer): PPixel;
    function GetHeight: Integer;
    function GetWidth: Integer;
  protected
    procedure Dead; virtual;
    procedure DrawAlphaSprite(aDest: TBitmap; aX, aY: Integer);
    procedure DrawNormalSprite(aDest: TBitmap; aX, aY: Integer);
    property Parent: TSprite read fParent write fParent;
    property AnimLoop: Boolean read fAnimLoop write fAnimLoop;
    property FireTimer: TTimer read fFireTimer write fFireTimer;
    property AnimBmp: TAnimBmp read fAnimBmp;
    property AnimNr: Byte read fAnimNr write SetAnimNr;
    property AnimDefaultNr: Byte read fAnimDefaultNr write fAnimDefaultNr;
    property FrameNr: Byte read fFrameNr write fFrameNr;
    property CenterX: Single read GetCenterX;
    property CenterY: Single read GetCenterY;
  public
    constructor Create(aParent: TSprite; aFileName: String; aLayer: TLayer; aX, aY: Single; aSpeed: Single = 0);
    destructor Destroy; override;
    procedure Move(aMoveCount: Single; aKeys: TKeys; aMouse: TMouse); virtual;
    procedure Draw(aFrame: TBitmap); virtual;
    procedure Hit(aSprite: TSprite); virtual; //abstract;
    function AllowHit(aSprite: TSprite): Boolean; virtual;
    function ClipRect: TRect; virtual;
    property isDead: Boolean read fDead;
    property Layer: TLayer read fLayer write fLayer;
    property Speed: Single read fSpeed write fSpeed;
    property X: Single read fX write fX;
    property Y: Single read fY write fY;
    property Width: Integer read GetWidth;
    property Height: Integer read GetHeight;
    property Pixel[aX, aY: Integer]: PPixel read GetPixel;
  end;

implementation

{ TSprite }

constructor TSprite.Create(aParent: TSprite; aFileName: String; aLayer: TLayer; aX, aY: Single; aSpeed: Single = 0);
begin
  inherited Create;
  fParent:=aParent;
  fAnimBmp:=GetBitmap(aFileName);
  fLayer:=aLayer;
  fX:=aX;
  fY:=aY;
  fSpeed:=aSpeed;
  fAnimLoop:=True;
  fAnimDefaultNr:=0;
  fAnimNr:=0;
  fFrameNr:=0;
  fFireTimer:=nil;
  fAnimTimer:=TTimer.Create(1000 div fAnimBmp.FPS[0]);
end;

destructor TSprite.Destroy;
begin
  FreeAndNil(fAnimTimer);
  if Assigned(fFireTimer) then
    FreeAndNil(fFireTimer);
  inherited Destroy;
end;

procedure TSprite.Move(aMoveCount: Single; aKeys: TKeys; aMouse: TMouse);
begin
  if fAnimTimer.Fire(aMoveCount) then begin
    Inc(fFrameNr);
    if fFrameNr=fAnimBmp.FrameCount[fAnimNr] then
      if not fAnimLoop then
        Dead
      else begin
        fAnimNr:=fAnimDefaultNr;
        fFrameNr:=0;
        fAnimTimer.Reset(1000 div fAnimBmp.FPS[0]);
      end;
  end;
end;

procedure TSprite.Draw(aFrame: TBitmap);
begin
  DrawAlphaSprite(aFrame, Round(fX), Round(fY));
end;

procedure TSprite.DrawAlphaSprite(aDest: TBitmap; aX, aY: Integer);
const
  bitsR = $F800;  // 1111100000000000
  bitsG = $07E0;  // 0000011111100000
  bitsB = $001F;  // 0000000000011111
var
  s   : PPixel;
  d   : PWord;
  Rect: TRect;
  x1  : Integer;
  y1  : Integer;
begin
  if (aX>=aDest.Width) or (aX+Pred(AnimBmp.Width)<0) or
     (aY>=aDest.Height) or (aY+Pred(AnimBmp.Height)<0) then exit;
  with Rect do begin
    Left:=Max(aX, 0);
    Top:=Max(aY, 0);
    Right:=Min(aX+Pred(AnimBmp.Width), Pred(aDest.Width));
    Bottom:=Min(aY+Pred(AnimBmp.Height), Pred(aDest.Height));
  end;
  for y1:=Rect.Top to Rect.Bottom do begin
    d:=aDest.ScanLine[y1];
    Inc(d, Rect.Left);
    s:=AnimBmp.Pixel[AnimNr, FrameNr, Rect.Left-aX, y1-aY];
    for x1:=Rect.Left to Rect.Right do begin
      if s^.alpha<>0 then
        d^:=(s^.color and bitsB * s^.alpha + d^ and bitsB * not s^.alpha) shr 8 or
            (s^.color and bitsG * s^.alpha + d^ and bitsG * not s^.alpha) shr 8 and bitsG or
            (s^.color and bitsR * s^.alpha + d^ and bitsR * not s^.alpha) shr 8 and bitsR;
      Inc(s);
      Inc(d);
    end;
  end;
end;

procedure TSprite.DrawNormalSprite(aDest: TBitmap; aX, aY: Integer);
var
  s   : PPixel;
  d   : PWord;
  Rect: TRect;
  x1  : Integer;
  y1  : Integer;
begin
  if (aX>=aDest.Width) or (aX+Pred(AnimBmp.Width)<0) or
     (aY>=aDest.Height) or (aY+Pred(AnimBmp.Height)<0) then exit;
  with Rect do begin
    Left:=Max(aX, 0);
    Top:=Max(aY, 0);
    Right:=Min(aX+Pred(AnimBmp.Width), Pred(aDest.Width));
    Bottom:=Min(aY+Pred(AnimBmp.Height), Pred(aDest.Height));
  end;
  for y1:=Rect.Top to Rect.Bottom do begin
    d:=aDest.ScanLine[y1];
    Inc(d, Rect.Left);
    s:=AnimBmp.Pixel[AnimNr, FrameNr, Rect.Left-aX, y1-aY];
    for x1:=Rect.Left to Rect.Right do begin
      d^:=s^.color;
      Inc(s);
      Inc(d);
    end;
  end;
end;

function TSprite.GetCenterX: Single;
begin
  Result:=fX+AnimBmp.Width/2;
end;

function TSprite.GetCenterY: Single;
begin
  Result:=fY+AnimBmp.Height/2;
end;

function TSprite.GetHeight: Integer;
begin
  Result:=fAnimBmp.Height;
end;

function TSprite.GetPixel(aX, aY: Integer): PPixel;
begin
  Result:=fAnimBmp.Pixel[AnimNr, fFrameNr, aX, aY];
end;

function TSprite.GetWidth: Integer;
begin
  Result:=fAnimBmp.Width;
end;

procedure TSprite.Hit(aSprite: TSprite);
begin

end;

function TSprite.AllowHit(aSprite: TSprite): Boolean;
begin
  Result:=False;
end;

procedure TSprite.Dead;
begin
  fDead:=True;
end;

function TSprite.ClipRect: TRect;
begin
  with Result do begin
    Left:=Round(fX);
    Top:=Round(fY);
    Right:=Left+AnimBmp.Width;
    Bottom:=Top+AnimBmp.Height;
  end;
end;

procedure TSprite.SetAnimNr(aAnimNr: Byte);
begin
  fAnimNr:=aAnimNr;
  fFrameNr:=0;
  fAnimTimer.Reset(1000 div fAnimBmp.FPS[fAnimNr]);
end;

end.
